import java.util.ArrayList;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.geometry.Bounds;
import javafx.geometry.BoundingBox;

/**
 * ロック
 * <pre>
 * ロックをする仕組みを担当します。
 * ロックした一覧を保持し、必要に応じ、敵スプライトの払い戻しを行います。
 * 時間が経ちすぎた場合、敵スプライトがロック解除必要な場合（範囲を外れノード削除／爆発した等）、
 * ロックを解除します。
 *
 * どこかのクラスに組み込もうと思ってましたが、
 * 思ったよりややこしくて、シンプルな作りを目指すため、
 * 単独のクラスと作ることとしました。
 * </pre>
 */
public class PRLock {

	/** アプリケーション */
	private		PRApplication				application			= null;

	/** ロックオンスプライト一覧 */
	private		ArrayList<PRSpriteT>		spriteLockLst		= null;

	/** 時間一覧 */
	private		ArrayList<int[]>			timeLst				= null;

	/** ロックオン対象外スプライト一覧 */
	private		ArrayList<PRSpriteT>		spriteLockNoLst		= null;

	/** 時間対象外一覧 */
	private		ArrayList<int[]>			timeNoLst			= null;

	/**
	 * コンストラクタ
	 * @param application アプリケーション
	 */
	public PRLock(PRApplication application){

		// アプリケーション保持
		this.application = application;

		// ロックオンスプライト一覧を生成
		spriteLockLst = new ArrayList<PRSpriteT>(100);

		// 時間一覧を生成
		timeLst = new ArrayList<int[]>(100);

		// ロックオン対象外スプライト一覧を生成
		spriteLockNoLst = new ArrayList<PRSpriteT>(100);

		// 時間対象外一覧を生成
		timeNoLst = new ArrayList<int[]>(100);

	} // end PRLock

	/**
	 * ロックオン
	 * <pre>
	 * ロックオンします。
	 * 1/6秒毎にチェックし、スプライト一覧に登録されている最初の方からロックするかチェックします。
	 * </pre>
	 */
	public void lockon(){

		// プレイヤーを取得
		PRSpritePrusher spritePrusher = this.application.getPRSpritePrusher();

		// 境界を取得
		Bounds boundsPrusher = spritePrusher.getBoundsInParent();

		// ロックオン範囲を設定
		Bounds boundsLock = new BoundingBox(
			boundsPrusher.getMinX() - 0.02,
			boundsPrusher.getMinY() - 0.02,
			boundsPrusher.getMinZ(),
			boundsPrusher.getWidth() + 0.04,
			boundsPrusher.getHeight() + 0.04,
			boundsPrusher.getDepth() + 100.0
		);

		// スプライト一覧を取得
		ObservableList<Node> nodeLst = this.application.getNodeLst();

		// スプライト一覧の件数を取得
		int countNodeLst = nodeLst.size();

		// スプライト一覧の件数分、ループ
		for(int i = 0; i < countNodeLst; i++){

			// スプライトを取得
			Node node = nodeLst.get(i);

			// 敵スプライトではない場合、
			if(!(node instanceof PRSpriteT)){
				// コンティニュー
				continue;
			}

			// 敵が低木の場合、
			if(node instanceof PRSpriteShrub){
				// コンティニュー
				continue;
			}

			// 敵が木の場合、
			if(node instanceof PRSpriteTree){
				// コンティニュー
				continue;
			}

			// 柱の場合、
			if(node instanceof PRSpritePole){
				// コンティニュー
				continue;
			}

			// ビルの場合、
			if(node instanceof PRSpriteBuilding){
				// コンティニュー
				continue;
			}

			// 雲の場合、
			if(node instanceof PRSpriteCloud){
				// コンティニュー
				continue;
			}

			// 既にロックオンされている場合、
			if(spriteLockLst.indexOf(node) != -1){
				// コンティニュー
				continue;
			}

			// ロックオン対象外スプライト一覧に存在する場合、
			if(spriteLockNoLst.indexOf(node) != -1){
				// コンティニュー
				continue;
			}

			// ロックオン範囲の場合、
			if(boundsLock.intersects(node.getBoundsInParent())){

				// 敵スプライトにキャスト
				PRSpriteT spriteT = (PRSpriteT)node;

				// ロックオン。ロックオンスプライト一覧に追加
				spriteLockLst.add(spriteT);

				// 時間一覧に追加
				timeLst.add(new int[]{6});

				// ロック音スタート
				this.application.startWaveLockon();

				// ループを抜ける
				break;

			} // end if ロックオン範囲の場合

		} // end for スプライト一覧の件数分、ループ

		// ロックオンスプライト一覧が100件を超えた場合、
		if(spriteLockLst.size() > 100){

			// ロックオンスプライト一覧（ワーク）を生成
			ArrayList<PRSpriteT> spriteLockWkLst = new ArrayList<PRSpriteT>(100);

			// 時間一覧（ワーク）を生成
			ArrayList<int[]> timeWkLst = new ArrayList<int[]>(100);

			// ロックオンスプライト一覧の件数を取得
			int countSpriteLockLst = spriteLockLst.size();

			// ロックオンスプライト一覧の件数分、ループ
			for(int i = 0; i < countSpriteLockLst; i++){

				// ロックオンスプライトを取得
				PRSpriteT spriteLock = this.spriteLockLst.get(i);

				// null以外の場合、
				if(spriteLock != null){

					// ロックオンスプライト一覧（ワーク）に追加
					spriteLockWkLst.add(spriteLock);

					// 時間を取得
					int[] times = this.timeLst.get(i);

					// 時間一覧（ワーク）に追加
					timeWkLst.add(times);

				} // end if null以外の場合

			} // end for ロックオンスプライト一覧の件数分、ループ

			// ロックオンスプライト一覧を入れ替え
			this.spriteLockLst = spriteLockWkLst;

			// 時間一覧を入れ替え
			this.timeLst = timeWkLst;

		} // end if ロックオンスプライト一覧が100件を超えた場合

	} // end lockon

	/**
	 * ロックを外す
	 * <pre>
	 * 時間が経ちすぎた場合、ロックを解除します。
	 * 1/6毎に処理を行います。
	 * </pre>
	 */
	public void unlock() {

		// ロックオンスプライト一覧の件数を取得
		int countSpriteLockLst = spriteLockLst.size();

		// ロックオンスプライト一覧の件数分、ループ
		for(int i = 0; i < countSpriteLockLst;i++){

			// 時間を取得
			int[] times = timeLst.get(i);

			// nullの場合、
			if(times == null){
				// コンティニュー
				continue;
			}

			// 時間−１
			times[0]--;

			// 時間が0になった場合、
			if(times[0] <= 0){

				// ロックオンスプライトを取得
				PRSpriteT spriteLock = this.spriteLockLst.get(i);

				// ロックオンスプライト一覧から削除
				spriteLockLst.set(i, null);
				// 時間一覧から削除
				timeLst.set(i, null);

				// ロックオン対象外スプライト一覧に追加
				spriteLockNoLst.add(spriteLock);
				// 時間対象外一覧に追加
				timeNoLst.add(new int[]{6});

			}

		} // end for ロック一覧の件数分、ループ

		// ロックオン対象外スプライト一覧の件数を取得
		int countSpriteLockNoLst = spriteLockNoLst.size();

		// ロックオン対象外スプライト一覧の件数分、ループ
		for(int i = 0; i < countSpriteLockNoLst;i++){

			// 時間を取得
			int[] timeNos = timeNoLst.get(i);

			// nullの場合、
			if(timeNos == null){
				// コンティニュー
				continue;
			}

			// 時間−１
			timeNos[0]--;

			// 時間が0になった場合、
			if(timeNos[0] <= 0){

				// ロックオン対象外スプライト一覧から削除
				spriteLockNoLst.set(i, null);
				// 時間対象外一覧から削除
				timeNoLst.set(i, null);

			}

		} // end for ロック一覧の件数分、ループ

		// ロックオン対象外スプライト一覧が100件を超えた場合、
		if(countSpriteLockNoLst > 100){

			// ロックオン対象外スプライト一覧（ワーク）を生成
			ArrayList<PRSpriteT> spriteLockNoWkLst = new ArrayList<PRSpriteT>(100);

			// 時間対象外一覧（ワーク）を生成
			ArrayList<int[]> timeNoWkLst = new ArrayList<int[]>(100);

			// ロックオン対象外スプライト一覧の件数分、ループ
			for(int i = 0; i < countSpriteLockNoLst; i++){

				// ロックオンスプライトを取得
				PRSpriteT spriteLockNo = this.spriteLockNoLst.get(i);

				// null以外の場合、
				if(spriteLockNo != null){

					// ロックオン対象外スプライト一覧（ワーク）に追加
					spriteLockNoWkLst.add(spriteLockNo);

					// 時間を取得
					int[] timeNos = this.timeNoLst.get(i);

					// 時間対象外一覧（ワーク）に追加
					timeNoWkLst.add(timeNos);

				} // end if null以外の場合

			} // end for ロックオン対象外スプライト一覧の件数分、ループ

			// ロックオン対象外スプライト一覧を入れ替え
			this.spriteLockNoLst = spriteLockNoWkLst;

			// 時間対象外一覧を入れ替え
			this.timeNoLst = timeNoWkLst;

		} // end if ロックオン対象外スプライト一覧が100件を超えた場合

	} // end unlock

	/**
	 * ノード削除
	 * <pre>
	 * ロック対象になっていた場合は外します。
	 * </pre>
	 */
	public void removeNode(Node node){

		// 敵スプライトではない場合、
		if(!(node instanceof PRSpriteT)){
			// リターン
			return;
		}

		// インデックスを取得
		int index = spriteLockLst.indexOf(node);

		// ロックオンされている場合、
		if(index != -1){

			// ロックオンスプライト一覧から削除
			spriteLockLst.set(index, null);
			// 時間一覧から削除
			timeLst.set(index, null);

		}

		// インデックスを取得
		int indexNo = spriteLockNoLst.indexOf(node);

		// ロックオン対象外となっている場合、
		if(indexNo != -1){

			// ロックオン対象外スプライト一覧から削除
			spriteLockNoLst.set(indexNo, null);
			// 時間対象外一覧から削除
			timeNoLst.set(indexNo, null);

		}

	} // end removeNode

	/**
	 * ロックオンしている敵スプライトを取得
	 * <pre>
	 * ロックオンスプライト一覧に登録されている最初の方からロックオンされている敵スプライトを返します。
	 * </pre>
	 * @return PRSpriteT ロックオンしている敵スプライト。ロックオンしていない場合は、nullをリターン。
	 */
	public PRSpriteT getLockOnSpriteT(){

		// ロックオンスプライト一覧の件数を取得
		int countSpriteLockLst = spriteLockLst.size();

		// ロックオンスプライト一覧の件数分、ループ
		for(int i = 0; i < countSpriteLockLst; i++){

			// ロックオンスプライトを取得
			PRSpriteT spriteLock = this.spriteLockLst.get(i);

			// null以外の場合、
			if(spriteLock != null){

				// ロックオンスプライト一覧から削除
				spriteLockLst.set(i, null);
				// 時間一覧から削除
				timeLst.set(i, null);

				// ロックオン対象外スプライト一覧に追加
				spriteLockNoLst.add(spriteLock);
				// 時間対象外一覧に追加
				timeNoLst.add(new int[]{12});

				// ロックオンスプライトをリターン
				return spriteLock;

			} // end if null以外の場合

		} // end for ロックオンスプライト一覧の件数分、ループ

		// nullでリターン
		return null;

	} // end getLockOnSpriteT

}
