import java.util.List;
import java.util.ArrayList;
import java.util.HashMap;

import javafx.collections.ObservableList;
import javafx.scene.Group;
import javafx.scene.Node;

/**
 * 3Dパネルクラス
 * <pre>
 * 時間の制御を行い、適宜、ノードのrunメソッドを呼び出します。
 * キーの押下状態を制御します。
 * </pre>
 */
public class PRPanel3D extends Group implements PRConstant {

	/** アプリケーション */
	private		PRApplication			application				= null;

	/** 影一覧 */
	private		HashMap<Node, Node>		mapShadow				= null;

	/**
	 * コンストラクタ
	 * @param application アプリケーション
	 */
	public PRPanel3D(PRApplication application){

		// スーパークラスを呼び出す
		super();

		// アプリケーション保持
		this.application = application;

		// 初期化処理を呼び出す
		init();

	} // end PRPanel3D

	/**
	 * 初期化処理
	 * <pre>
	 * パネルクラスの初期化処理を行います。
	 * </pre>
	 */
	public void init(){

		// 影マップを生成
		mapShadow = new HashMap<Node, Node>();

	} // end init

	/**
	 * 開始
	 * <pre>
	 * 当オブジェクトの処理を開始します。
	 * </pre>
	 */
	public void start(){

	}

	/**
	 * 実行
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void run(){

		// アプリケーションの実行メソッド呼び出し
		boolean flgPlaying = application.isPlaying();

		// ゲーム中の場合、
		if(flgPlaying){

			//
			// ノード処理
			//

			// ノード一覧を取得
			List<Node> nodeLst = this.getNodeLst();

			// ノード一覧の件数を取得
			int countNodeLst = nodeLst.size();

			// ノード一覧の件数分、ループ
			for(int j = 0; j < countNodeLst; j++){

				// ノードを取得
				Node node = nodeLst.get(j);

				// ダミーノードの場合、
				if(node instanceof PRNodeDummy){

					// コンティニュー
					continue;

				// 影の場合、
				}else if(node instanceof PRSpriteShadow){

					// コンティニュー
					continue;

				// スプライトの場合、
				}else if(node instanceof PRSprite){

					// スプライトを取得
					PRSprite sprite = (PRSprite)node;

					// 衝突中か？
					boolean flgBump = application.isBump();

					// 衝突中 かつ プレイヤーではない場合、
					if(flgBump && !(sprite instanceof PRSpritePrusher)){

						// コンティニュー
						continue;

					}

					// スプライトを実行
					sprite.run();

					// 影がある場合、
					if(mapShadow.containsKey(sprite)){

						// 影を取得
						PRSpriteShadow spriteShadow = (PRSpriteShadow)mapShadow.get(sprite);

						// スプライトを実行
						spriteShadow.run();

					}

				} // end if スプライトの場合

			} // end for ノード一覧の件数分、ループ

		} // end if ゲーム中の場合（ポーズ中ではない場合）

	} // end run

	/**
	 * お掃除の時間
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void clean(){

		//
		// ノード一覧が多くなった場合は、お掃除をする。（メモリ増加防止）
		// ノードの実行処理とは別にする（処理分散）
		//

		// ノード一覧を取得
		ObservableList<Node> nodeLst = this.getNodeLst();

		// ノード一覧の件数を取得
		int countNodeLst = nodeLst.size();

		// ノード一覧の件数が1000件を超えた場合、
		if(countNodeLst > 1000){

			// 一時的にノード一覧（ワーク）を作成
			ArrayList<Node> nodeLstWk = new ArrayList<Node>(1000);

			// ノード一覧の件数分、ループ
			for(int j = 0; j < countNodeLst; j++){

				// ノードを取得
				Node node = (Node)nodeLst.get(j);

				// ノードがダミーノード以外の場合、
				if(!(node instanceof PRNodeDummy)){

					// ノード一覧（ワーク）に追加
					nodeLstWk.add(node);

				}

			} // end for ノード一覧の件数分、ループ

			// ノード一覧を入れ替える
			nodeLst.setAll(nodeLstWk);

		} // end if ノード一覧の件数が1000件を超えた場合

	} // end clean

	/**
	 * ノード追加
	 * @param node ノード
	 */
	public void addNode(Node node){

		// ノード一覧を取得
		List<Node> nodeLst = this.getNodeLst();

		// ノードを追加
		nodeLst.add(node);

	} // end addNode

	/**
	 * ノード削除
	 * <pre>
	 * ノードを削除します。
	 * 当メソッドでは、ダミーノードに置き換え、runメソッドの中で削除します。
	 * nullに置き換えができない
	 * かつ Listクラスにremoveメソッドを多用することで配列コピーを多用されることを避けるため
	 * （重くなることをできるだけ避ける）
	 * </pre>
	 * @param node ノード
	 */
	public void removeNode(Node node){

		// ノード一覧を取得
		List<Node> nodeLst = this.getNodeLst();

		// インデックスを求める
		int index = nodeLst.indexOf(node);

		// インデックスが取得できた場合、
		if(index != -1){

			// ダミーノードに置き換え
			nodeLst.set(index, new PRNodeDummy(this.application));

		}

		// 影ノード削除
		// 影が存在する場合のみ削除する
		this.removeShadow(node);

	} // end removeNode

	/**
	 * 影ノード追加
	 * <pre>
	 * 影ノードを追加します。
	 * </pre>
	 * @param sprite ノード
	 */
	public void addShadow(Node node){

		// 影ノード追加
		this.addShadow(node, -1);

	}

	/**
	 * 影ノード追加
	 * <pre>
	 * 影ノードを追加します。
	 * </pre>
	 * @param sprite ノード
	 * @param radius 半径
	 */
	public void addShadow(Node node, double radius){

		// 影を生成
		PRSpriteShadow spriteShadow = new PRSpriteShadow(application, node, radius);

		// 3Dパネルクラスを呼び出し
		this.addNode(spriteShadow);

		// 影を紐付け
		mapShadow.put(node, spriteShadow);

	}

	/**
	 * 影ノード削除
	 * <pre>
	 * 影ノードを削除します。
 	 * 影が存在する場合のみ削除します。
 	 * 影が存在しない場合は何もしません（エラーにはなりません）。
	 * </pre>
	 * @param sprite ノード
	 */
	public void removeShadow(Node node){

		// ノード一覧を取得
		List<Node> nodeLst = this.getNodeLst();

		// 影が存在する場合、
		if(mapShadow.containsKey(node)){

			// インデックスを求める
			int index = nodeLst.indexOf(mapShadow.get(node));

			// インデックスが取得できた場合、
			if(index != -1){

				// 影一覧から削除
				// ※※※ 入れ忘れると、メモリーリークとなる。
				// 影オブジェクトが残ったままとなり、メモリを食いつぶすことになる。
				mapShadow.remove(node);

				// ダミーノードに置き換え
				nodeLst.set(index, new PRNodeDummy(this.application));

			}

		} // end if 影が存在する場合

	} // end removeShadow

	/**
	 * ノード一覧取得
	 * @return ObservableList<Node> ノード一覧
	 */
	public ObservableList<Node> getNodeLst(){

		// Listを取得
		ObservableList<Node> nodeLst = this.getChildren();

		// リターン
		return nodeLst;

	} // end getNodeLst

	/**
	 * アプリケーション取得
	 * @return PRApplication アプリケーション
	 */
	public PRApplication getApplication(){

		// アプリケーションをリターン
		return application;

	} // end getApplication

}
